/**
 * EasyBeans
 * Copyright (C) 2006 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:EjbJarClassMetadata.java 2372 2008-02-08 18:18:37Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.deployment.metadata.ejbjar;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import org.ow2.easybeans.deployment.annotations.exceptions.InterceptorsValidationException;
import org.ow2.util.ee.deploy.api.deployable.EJB3Deployable;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarClassMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarFieldMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarMethodMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.InterceptorType;
import org.ow2.util.ee.metadata.ejbjar.impl.specific.SpecificEjbJarClassMetadata;
import org.ow2.util.pool.api.IPoolConfiguration;
import org.ow2.util.pool.api.IPoolMetadata;

/**
 * This class represents the annotation metadata of a Bean.<br>
 * From this class, we can access to all methods of a bean with its associated
 * information.
 * @author Florent Benoit
 */
public class EjbJarClassMetadata extends
        SpecificEjbJarClassMetadata<EJB3Deployable, EjbJarClassMetadata, EjbJarMethodMetadata, EjbJarFieldMetadata> implements
        IPoolMetadata {

    private EjbJarArchiveMetadata ejbJarArchiveMetadata;

    /**
     * List of inherited interfaces (present in the super classes).
     */
    private List<String> inheritedInterfaces = null;

    /**
     * List of &#64;{@link javax.interceptor.AroundInvoke} methods on this
     * class (should be only one per class, validating occurs after).
     */
    private List<EjbJarMethodMetadata> aroundInvokeMethodsMetadata = null;

    /**
     * Methods used for &#64;{@link javax.annotation.PostConstruct} on this
     * class (only one per class but may be defined in super classes).
     */
    private LinkedList<EjbJarMethodMetadata> postConstructMethodsMetadata = null;

    /**
     * Methods used for &#64;{@link javax.annotation.PreDestroy} on this class
     * (only one per class but may be defined in super classes).
     */
    private LinkedList<EjbJarMethodMetadata> preDestroyMethodsMetadata = null;

    /**
     * Methods used for &#64;{@link javax.ejb.PostActivate} on this class (only
     * one per class but may be defined in super classes).
     */
    private LinkedList<EjbJarMethodMetadata> postActivateMethodsMetadata = null;

    /**
     * Methods used for &#64;{@link javax.ejb.PrePassivate} on this class (only
     * one per class but may be defined in super classes).
     */
    private LinkedList<EjbJarMethodMetadata> prePassivateMethodsMetadata = null;

    /**
     * Pool's configuration.
     */
    private IPoolConfiguration poolConfiguration;

    /**
     * Cluster's configuration.
     */
    private Object cluster;

    public EjbJarClassMetadata(final IEjbJarClassMetadata<EJB3Deployable> warClassMetadata) {
        super(warClassMetadata);
        this.postConstructMethodsMetadata = new LinkedList<EjbJarMethodMetadata>();
        this.preDestroyMethodsMetadata = new LinkedList<EjbJarMethodMetadata>();
        this.postActivateMethodsMetadata = new LinkedList<EjbJarMethodMetadata>();
        this.prePassivateMethodsMetadata = new LinkedList<EjbJarMethodMetadata>();
        this.inheritedInterfaces = new ArrayList<String>();
    }

    public EjbJarArchiveMetadata getEjbJarArchiveMetadata() {
        return this.ejbJarArchiveMetadata;
    }

    /**
     * @return the method metadata with annotation &#64;{@link javax.interceptor.AroundInvoke}.
     */
    public boolean isAroundInvokeMethodMetadata() {
        return (this.aroundInvokeMethodsMetadata != null);
    }

    /**
     * @return the list of methods metadata with annotation &#64;{@link javax.interceptor.AroundInvoke}.
     */
    public List<EjbJarMethodMetadata> getAroundInvokeMethodMetadatas() {
        return this.aroundInvokeMethodsMetadata;
    }

    /**
     * Add a &#64;{@link javax.interceptor.AroundInvoke} method of this class.
     * @param aroundInvokeMethodMetadata the method.
     */
    public void addAroundInvokeMethodMetadata(final EjbJarMethodMetadata aroundInvokeMethodMetadata) {
        if (this.aroundInvokeMethodsMetadata == null) {
            this.aroundInvokeMethodsMetadata = new ArrayList<EjbJarMethodMetadata>();
        }
        this.aroundInvokeMethodsMetadata.add(aroundInvokeMethodMetadata);
    }

    /**
     * @return the methods metadata with annotation &#64;{@link javax.annotation.PostConstruct}.
     */
    public LinkedList<EjbJarMethodMetadata> getPostConstructMethodsMetadata() {
        return this.postConstructMethodsMetadata;
    }

    /**
     * Adds a &#64;{@link javax.annotation.PostConstruct} method of this class.
     * @param postConstructMethodMetadata the method.
     */
    public void addPostConstructMethodMetadata(final EjbJarMethodMetadata postConstructMethodMetadata) {
        checkLifeCycleDuplicate(postConstructMethodMetadata, InterceptorType.POST_CONSTRUCT, getPostConstructMethodsMetadata());
        this.postConstructMethodsMetadata.addFirst(postConstructMethodMetadata);
    }

    /**
     * Checks that only method at one level of a class is present.
     * @param postConstructMethodMetadata method to check
     * @param itcType the type of interceptor (used for the error)
     * @param existingList current list of methods
     */
    private void checkLifeCycleDuplicate(final EjbJarMethodMetadata postConstructMethodMetadata,
            final InterceptorType itcType, final List<EjbJarMethodMetadata> existingList) {
        // FIXME gaellalire : setInherit => specific to ezb
        // First case : not inherited
        IEjbJarClassMetadata<?> wantToAddClassMetadata = postConstructMethodMetadata.getSpecificClassMetadata();
        if (postConstructMethodMetadata.isInherited()) {
            wantToAddClassMetadata = postConstructMethodMetadata.getOriginalClassMetadata();
        }
        for (EjbJarMethodMetadata method : existingList) {
            IEjbJarClassMetadata<?> compareMetaData;
            if (method.isInherited()) {
                compareMetaData = method.getOriginalClassMetadata();
            } else {
                compareMetaData = method.getClassMetadata();
            }
            if (compareMetaData.equals(wantToAddClassMetadata)) {
                throw new InterceptorsValidationException("Class " + getClassName() + " has already a " + itcType
                        + " method which is " + method.getMethodName() + ", cannot set new method "
                        + postConstructMethodMetadata.getMethodName());
            }
        }
    }

    /**
     * @return the methods metadata with annotation &#64;{@link javax.annotation.PreDestroy}.
     */
    public LinkedList<EjbJarMethodMetadata> getPreDestroyMethodsMetadata() {
        return this.preDestroyMethodsMetadata;
    }

    /**
     * Adds a &#64;{@link javax.annotation.PreDestroy} method of this class.
     * @param preDestroyMethodMetadata the method.
     */
    public void addPreDestroyMethodMetadata(final EjbJarMethodMetadata preDestroyMethodMetadata) {
        checkLifeCycleDuplicate(preDestroyMethodMetadata, InterceptorType.PRE_DESTROY, getPreDestroyMethodsMetadata());
        this.preDestroyMethodsMetadata.addFirst(preDestroyMethodMetadata);
    }

    /**
     * @return the methods metadata with annotation &#64;{@link javax.ejb.PostActivate}.
     */
    public LinkedList<EjbJarMethodMetadata> getPostActivateMethodsMetadata() {
        return this.postActivateMethodsMetadata;
    }

    /**
     * Adds a &#64;{@link javax.ejb.PostActivate} method of this class.
     * @param postActivateMethodMetadata the method.
     */
    public void addPostActivateMethodMetadata(final EjbJarMethodMetadata postActivateMethodMetadata) {
        checkLifeCycleDuplicate(postActivateMethodMetadata, InterceptorType.POST_ACTIVATE, getPostActivateMethodsMetadata());
        this.postActivateMethodsMetadata.addFirst(postActivateMethodMetadata);
    }

    /**
     * @return the method metadata with annotation &#64;{@link javax.ejb.PrePassivate}.
     */
    public LinkedList<EjbJarMethodMetadata> getPrePassivateMethodsMetadata() {
        return this.prePassivateMethodsMetadata;
    }

    /**
     * Adds a &#64;{@link javax.ejb.PrePassivate} method of this class.
     * @param prePassivateMethodMetadata the method.
     */
    public void addPrePassivateMethodMetadata(final EjbJarMethodMetadata prePassivateMethodMetadata) {
        checkLifeCycleDuplicate(prePassivateMethodMetadata, InterceptorType.PRE_PASSIVATE, getPrePassivateMethodsMetadata());
        this.prePassivateMethodsMetadata.addFirst(prePassivateMethodMetadata);
    }

    /**
     * Is that this class is an interceptor class ?
     * @return true if it the case, else false.
     */
    public boolean isInterceptor() {
        return (this.aroundInvokeMethodsMetadata != null && this.aroundInvokeMethodsMetadata.size() > 0)
                || (this.postConstructMethodsMetadata != null && this.postConstructMethodsMetadata.size() > 0)
                || (this.preDestroyMethodsMetadata != null && this.preDestroyMethodsMetadata.size() > 0)
                || (this.prePassivateMethodsMetadata != null && this.prePassivateMethodsMetadata.size() > 0)
                || (this.postActivateMethodsMetadata != null && this.postActivateMethodsMetadata.size() > 0);
    }

    /**
     * @return the pool configuration for this metadata.
     */
    public IPoolConfiguration getPoolConfiguration() {
        return this.poolConfiguration;
    }

    /**
     * Sets the pool configuration for this metadata.
     * @param poolConfiguration the given configuration.
     */
    public void setPoolConfiguration(final IPoolConfiguration poolConfiguration) {
        this.poolConfiguration = poolConfiguration;
    }

    public void setSpecificDeployableMetadata(final EjbJarArchiveMetadata warDeployableMetadata) {
        this.ejbJarArchiveMetadata = warDeployableMetadata;
    }

    @Override
    public void addStandardMethodMetadata(final IEjbJarMethodMetadata<EJB3Deployable> methodAnnotationMetadata) {
        throw new IllegalStateException("please use add Specific method metadata");
    }

    @Override
    public void addStandardFieldMetadata(final IEjbJarFieldMetadata<EJB3Deployable> fieldAnnotationMetadata) {
        throw new IllegalStateException("please use add Specific method metadata");
    }

    /**
     * @return the iherited interfaces of the super classes.
     */
    public List<String> getInheritedInterfaces() {
        return this.inheritedInterfaces;
    }

    /**
     * Sets the interfaces used in the super classes.
     * @param inheritedInterfaces the given interfaces
     */
    public void setInheritedInterfaces(final List<String> inheritedInterfaces) {
        this.inheritedInterfaces = inheritedInterfaces;
    }

    /**
     * Gets the cluster configuration
     * @return the cluster
     */
    public Object getCluster() {
        return this.cluster;
    }

    /**
     * Sets the cluster configuration.
     * @param cluster the clusterConfiguration to set
     */
    public void setCluster(final Object cluster) {
        this.cluster = cluster;
    }
}