/**
 * EasyBeans
 * Copyright (C) 2006-2007 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: EjbJarMethodMetadata.java 4200 2008-10-15 09:49:00Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.deployment.metadata.ejbjar;

import org.ow2.util.ee.deploy.api.deployable.EJB3Deployable;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarClassMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IEjbJarMethodMetadata;
import org.ow2.util.ee.metadata.ejbjar.api.IJClassInterceptor;
import org.ow2.util.ee.metadata.ejbjar.impl.specific.SpecificEjbJarMethodMetadata;
import org.ow2.util.scan.api.metadata.structures.JMethod;

import java.util.List;

/**
 * This class represents the annotation metadata of a method.
 * @author Florent Benoit
 */
public class EjbJarMethodMetadata extends
        SpecificEjbJarMethodMetadata<EJB3Deployable, EjbJarClassMetadata, EjbJarMethodMetadata, EjbJarFieldMetadata> implements
        Cloneable {

    /**
     * This method is a method from a super class ?<br>
     */
    private boolean inherited = false;

    /**
     * Original parent metadata (if method is inherited).
     */
    private EjbJarClassMetadata originalClassMetadata = null;

    /**
     * Class metadata.
     */
    private EjbJarClassMetadata classMetadata = null;

    /**
     * Method signature.
     */
    private JMethod method;

    /**
     * EasyBeans method interceptors. These interceptors correspond to a list of Interceptors like security or transaction.
     */
    private List<? extends IJClassInterceptor> interceptors = null;


    /**
     * Constructor.
     * @param warMethodMetadata standard war method metadata
     * @param classMetadata parent specific metadata
     */
    public EjbJarMethodMetadata(final IEjbJarMethodMetadata<EJB3Deployable> warMethodMetadata,
            final EjbJarClassMetadata classMetadata) {
        super(warMethodMetadata, classMetadata);
        this.method = warMethodMetadata.getJMethod();
    }

    /**
     * Another constructor (used by enhancer and resolver).
     * @param removeMethod the method signature
     * @param bean parent specific metadata
     */
    public EjbJarMethodMetadata(final JMethod removeMethod, final EjbJarClassMetadata bean) {
        super(new org.ow2.util.ee.metadata.ejbjar.impl.EjbJarMethodMetadata<EJB3Deployable>(removeMethod, bean.getStandard()),
                bean);
        this.method = removeMethod;
        // do not do : bean.getStandard().addMethodMetadata(getStandard());
        // because the standard should not be modified
    }

    /**
     * @return true if this method is inherited from a super class
     */
    public boolean isInherited() {
        return this.inherited;
    }

    /**
     * Sets the inheritance of this method.
     * @param inherited true if method is from a super class
     * @param originalClassMetadata the metadata of the original class (not
     *        inherited)
     */
    public void setInherited(final boolean inherited, final EjbJarClassMetadata originalClassMetadata) {
        this.inherited = inherited;
        this.originalClassMetadata = originalClassMetadata;
    }

    /**
     * @return original parent metadata (class) if inherited.
     */
    public EjbJarClassMetadata getOriginalClassMetadata() {
        return this.originalClassMetadata;
    }

    /**
     * @param classMetadata the new class metadata
     */
    public void setSpecificClassMetadata(final EjbJarClassMetadata classMetadata) {
        this.classMetadata = classMetadata;
    }

    /**
     * Prevent use of this method by throwing an {@link IllegalStateException}.
     * @param classMetadata standard parent class metadata
     */
    @Override
    public void setClassMetadata(final IEjbJarClassMetadata<EJB3Deployable> classMetadata) {
        throw new IllegalStateException("use setSpecificClassMetadata");
    }

    /**
     * @return the standard parent class metadata
     */
    @Override
    public IEjbJarClassMetadata<EJB3Deployable> getClassMetadata() {
        if (this.classMetadata == null) {
            return super.getClassMetadata();
        }
        return this.classMetadata.getStandard();
    }

    /**
     * @return the specific parent class metadata
     */
    @Override
    public EjbJarClassMetadata getSpecificClassMetadata() {
        if (this.classMetadata == null) {
            return super.getSpecificClassMetadata();
        }
        return this.classMetadata;
    }

    /**
     * Override because {@link org.ow2.easybeans.enhancer.interceptors.InterceptorClassAdapter} use it.
     * @param method the method signature
     */
    @Override
    public void setJMethod(final JMethod method) {
        this.method = method;
    }

    /**
     * Override because {@link org.ow2.easybeans.enhancer.interceptors.InterceptorClassAdapter} use it.
     * @return the method signature
     */
    @Override
    public JMethod getJMethod() {
        return this.method;
    }

    /**
     * Override because {@link org.ow2.easybeans.enhancer.interceptors.InterceptorClassAdapter} use it.
     * @return the method name
     */
    @Override
    public String getMethodName() {
        return getJMethod().getName();
    }

    /**
     * @return list of interceptors that enhancer will use. (ie : security/transaction)
     */
    public List<? extends IJClassInterceptor> getInterceptors() {
        return this.interceptors;
    }

    /**
     * Sets the list of interceptors(tx, security, etc) that enhancers will use.<br>
     * These interceptors are defined per methods.
     * @param interceptors list of interceptors that enhancer will use.
     */
    public void setInterceptors(final List<? extends IJClassInterceptor> interceptors) {
        this.interceptors = interceptors;
    }

    /**
     * @return a clone of this metadata.
     */
    @Override
    public Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException e) {
            return null;
        }
    }

}
