/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: MetadataSpecificMerge.java 4049 2008-10-03 12:51:24Z eyindanga $
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.deployment.metadata.ejbjar.helper;

import java.util.List;

import org.ow2.easybeans.deployment.metadata.ejbjar.EjbJarArchiveMetadata;
import org.ow2.easybeans.deployment.metadata.ejbjar.EjbJarClassMetadata;
import org.ow2.easybeans.deployment.metadata.ejbjar.xml.AbsSpecificBean;
import org.ow2.easybeans.deployment.metadata.ejbjar.xml.EJB;
import org.ow2.easybeans.deployment.metadata.ejbjar.xml.EasyBeansDD;
import org.ow2.easybeans.deployment.metadata.ejbjar.xml.MessageDrivenBean;
import org.ow2.easybeans.deployment.metadata.ejbjar.xml.Session;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.pool.api.IPoolConfiguration;

/**
 * Merge specific XML data into the metadata model.
 * @author Florent BENOIT
 */
public final class MetadataSpecificMerge {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(MetadataSpecificMerge.class);

    /**
     * ejb metadata.
     */
    private EjbJarArchiveMetadata ejbMetadata;

    /**
     * Helper class, no public constructor.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    private MetadataSpecificMerge(final EjbJarArchiveMetadata ejbMetadata) {
        this.ejbMetadata = ejbMetadata;
        // super(ejbMetadata);
    }

    /**
     * Takes struct of metadata and adds/set information on class/methods
     * metadata.
     * @param ejbMetadata the metadata corresponding to an EJB-JAR file.
     */
    public static void merge(final EjbJarArchiveMetadata ejbMetadata) {
        new MetadataSpecificMerge(ejbMetadata).resolve();
    }

    /**
     * Do all merging operations.
     */
    private void resolve() {
        EasyBeansDD easybeansDD = this.ejbMetadata.getEasyBeansDD();
        if (easybeansDD != null) {
            logger.debug("There is a specific Deployment Descriptor object, performing the merge of the metadata.");
            // Analyze EJBs
            EJB ejb = easybeansDD.getEJB();
            if (ejb != null) {

                // get session beans
                List<Session> sessionList = ejb.getSessions();
                if (sessionList != null) {
                    for (Session session : sessionList) {
                        // get ejb-name
                        String ejbName = session.getEjbName();
                        EjbJarClassMetadata classAnnotationMetadata = this.ejbMetadata
                                .getEjbJarClassMetadataForEjbName(ejbName);

                        // class is found, then apply all customization.
                        applySessionBean(session, classAnnotationMetadata);
                    }
                }

                // get MDB
                List<MessageDrivenBean> mdbList = ejb.getMessageDrivenBeans();
                if (mdbList != null) {
                    for (MessageDrivenBean mdb : mdbList) {
                        // get ejb-name
                        String ejbName = mdb.getEjbName();
                        EjbJarClassMetadata classAnnotationMetadata = this.ejbMetadata
                                .getEjbJarClassMetadataForEjbName(ejbName);

                        // class is found, then apply all customization.
                        applyMessageDrivenBean(mdb, classAnnotationMetadata);
                    }
                }
            }
        }
    }

    /**
     * Apply rules for a session bean.
     * @param sessionbean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applySessionBean(final Session sessionbean, final EjbJarClassMetadata classAnnotationMetadata) {
        applyCommonBean(sessionbean, classAnnotationMetadata);
    }

    /**
     * Apply rules for an MDB.
     * @param mdb the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyMessageDrivenBean(final MessageDrivenBean mdb, final EjbJarClassMetadata classAnnotationMetadata) {
        applyCommonBean(mdb, classAnnotationMetadata);
    }

    /**
     * Apply rules for a common bean.
     * @param bean the struct (XML DD)
     * @param classAnnotationMetadata (Annotation metadata)
     */
    private void applyCommonBean(final AbsSpecificBean bean, final EjbJarClassMetadata classAnnotationMetadata) {
        // Pool configured ? apply it
        IPoolConfiguration poolConfiguration = bean.getPoolConfiguration();
        Object clusterConfiguration = bean.getCluster();
        if (poolConfiguration != null) {
            classAnnotationMetadata.setPoolConfiguration(poolConfiguration);
        }
        if (clusterConfiguration != null) {
            classAnnotationMetadata.setCluster(clusterConfiguration);
        }
    }

}
