/**
 * EasyBeans
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:$
 * --------------------------------------------------------------------------
 */

package org.ow2.easybeans.injection.impl;

import org.ow2.easybeans.asm.Type;
import org.ow2.easybeans.injection.api.ArchiveInjectionException;
import org.ow2.easybeans.resolver.api.EZBJNDIData;
import org.ow2.easybeans.resolver.api.EZBRemoteJNDIResolver;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.CARDeployable;
import org.ow2.util.ee.deploy.api.deployable.IDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelper;
import org.ow2.util.ee.deploy.impl.helper.DeployableHelperException;
import org.ow2.util.ee.metadata.car.api.ICarDeployableMetadata;
import org.ow2.util.ee.metadata.car.impl.CarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.common.api.ICommonClassMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonFieldMetadata;
import org.ow2.util.ee.metadata.common.api.ICommonMethodMetadata;
import org.ow2.util.ee.metadata.common.api.interfaces.ISharedMetadata;
import org.ow2.util.ee.metadata.common.api.struct.IJAnnotationResource;
import org.ow2.util.ee.metadata.common.api.struct.IJEjbEJB;
import org.ow2.util.ee.metadata.common.api.struct.IJaxwsWebServiceRef;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.metadata.structures.JClass;
import org.ow2.util.scan.api.metadata.structures.JField;

import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import javax.xml.namespace.QName;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.rmi.RemoteException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * This class manages the injection
 * @author Florent Benoit
 */
public class ArchiveInjection {

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(ArchiveInjection.class);

    /**
     * Defines java.lang.Object class.
     */
    public static final String JAVA_LANG_OBJECT = "java/lang/Object";

    /**
     * Archive that will be used to analyze the classes.
     */
    private IArchive archive = null;

    /**
     * Analyze of the archive has been done ?
     */
    private boolean analyzed = false;

    /**
     * Remote JNDI Resolver (if present).
     */
    private EZBRemoteJNDIResolver jndiResolver = null;

    /**
     * Collection of metadata that have been analyzed.
     */
    private Collection<? extends ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata>> metadataCollection = null;

    /**
     * Default constructor.
     * @param archive the given archive on which injection is based.
     */
    public ArchiveInjection(final IArchive archive) {
        this.archive = archive;
    }

    /**
     * Analyze the archive.
     * @throws ArchiveInjectionException if analyze fails
     */
    public void analyze() throws ArchiveInjectionException {

        // Get Deployable
        IDeployable<?> deployable = null;
        try {
            deployable = DeployableHelper.getDeployable(archive);
        } catch (DeployableHelperException e) {
            throw new ArchiveInjectionException("Unable to get a deployable on archive '" + archive + "'.", e);
        }



        // Client Archive ?
        if (CARDeployable.class.isInstance(deployable)) {
            // Create metadata
            ICarDeployableMetadata carDeployableMetadata = null;
            try {
                carDeployableMetadata = new CarDeployableMetadataFactory().createDeployableMetadata(CARDeployable.class.cast(deployable));
            } catch (DeployableMetadataException e) {
                throw new ArchiveInjectionException("Unable to get metadata on archive '" + archive + "'.", e);
            }

            // Use only common part
            this.metadataCollection = carDeployableMetadata.getCarClassMetadataCollection();
        }  else {
            //TODO : adapt for other archives
            throw new UnsupportedOperationException("Can only manage injection for Client Deployable. Deployable found is '" + deployable + "'.");
        }

        // Ok, now that metadata are here,


        analyzed = true;
    }


    /**
     * Initialize stuff in the given class.
     * @param clazz the class to initialize
     */
    public void init(final Class<?> clazz) throws ArchiveInjectionException {
        // Analyze the metadata if not done
        if (!analyzed) {
            analyze();
        }

        // Get list of classes in the reverse order of the inheritance
        LinkedList<Class<?>> listClasses = new LinkedList<Class<?>>();
        Class<?> currentClass = clazz;
        while (currentClass != null) {
            listClasses.addFirst(currentClass);
            currentClass = currentClass.getSuperclass();
        }

        // Analyze classes and perform initialization
        for (Class<?> classToAnalyze : listClasses) {
            // Get the metadata of this class
            ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> classMetadata;
            classMetadata = getClassMetadata(classToAnalyze);

            // Found metadata ?
            if (classMetadata != null) {

                // Perform injection
                init(classMetadata, classToAnalyze, null);
            }
        }
    }

    /**
     * Perform injection in the given instance (or if null on a static way on the class) with the provided metadata
     * @param commonMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    public void init(ICommonClassMetadata commonMetadata, Class clazz, Object instance) throws ArchiveInjectionException {

        // Loop on the fields
       Collection<? extends ICommonFieldMetadata>fieldCollection = commonMetadata.getStandardFieldMetadataCollection();
       Iterator<? extends ICommonFieldMetadata> itField = fieldCollection.iterator();
        while (itField.hasNext()) {
            ICommonFieldMetadata fieldMetadata = itField.next();

            // Set the interface name by using the descriptor of the field
            Type typeInterface = Type.getType(fieldMetadata.getJField().getDescriptor());
            String interfaceName = typeInterface.getClassName();


            // Inject EJB
            IJEjbEJB jejb = fieldMetadata.getJEjbEJB();
            if (jejb != null) {
                // Update interface name
                jejb.setBeanInterface(interfaceName);
                // inject field
                injectFieldEJB(jejb, fieldMetadata, clazz, instance);
            }

            // Inject Resources
            IJAnnotationResource jAnnotationResource = fieldMetadata.getJAnnotationResource();
            if (jAnnotationResource != null) {
                // Update interface name
                jAnnotationResource.setType(interfaceName);

                // For some special case, fix the JNDI name to use
                if ("javax.transaction.UserTransaction".equals(jAnnotationResource.getType())) {
                    jAnnotationResource.setMappedName("java:comp/UserTransaction");
                } else if ("org.omg.CORBA.ORB".equals(jAnnotationResource.getType())) {
                    jAnnotationResource.setMappedName("java:comp/ORB");
                }

                // inject field
                injectFieldResource(jAnnotationResource, fieldMetadata, clazz, instance);
            }

            // Inject WebServiceRef
            IJaxwsWebServiceRef jaxwsWSR = fieldMetadata.getJaxwsWebServiceRef();
            if (jaxwsWSR != null) {

                // Update type name
                jaxwsWSR.setType(interfaceName);

                // Inject field
                injectWebServiceRef(jaxwsWSR, fieldMetadata, clazz, instance);
            }
        }

        // Visit metadatas at class-level
        List<IJaxwsWebServiceRef> webServiceRefs = commonMetadata.getJaxwsWebServiceRefs();
        if (webServiceRefs != null) {
            for (IJaxwsWebServiceRef ref : webServiceRefs) {

                // Inject class
                injectWebServiceRef(ref, commonMetadata, clazz, instance);

            }
        }

        // PostConstruct methods
        List<ICommonMethodMetadata> postConstructMethods = new ArrayList<ICommonMethodMetadata>();

        // Loop on the methods
        Collection<? extends ICommonMethodMetadata> methodCollection = commonMetadata.getStandardMethodMetadataCollection();
        Iterator<? extends ICommonMethodMetadata> itMethod = methodCollection.iterator();
         while (itMethod.hasNext()) {
             ICommonMethodMetadata methodMetadata = itMethod.next();

             // PostConstruct Method ?
             if (methodMetadata.isPostConstruct()) {
                postConstructMethods.add(methodMetadata);
             }

         }

        // Call Post Construct methods
        invokePostConstructMethods(postConstructMethods, clazz, instance);
    }

    /**
     * Perform injection in the given instance (or if null on a static way on the class) with
     * the provided metadata.
     * @param jaxwsWSR the service-ref that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    private void injectWebServiceRef(final IJaxwsWebServiceRef jaxwsWSR,
                                     final ISharedMetadata fieldMetadata,
                                     final Class clazz,
                                     final Object instance) throws ArchiveInjectionException {

        if (isNullOrEmpty(jaxwsWSR.getValue())) {
            // if the value was not set, we use the type instead
            jaxwsWSR.setValue(jaxwsWSR.getType());
        }

        // The service class is in 'value'
        ClassLoader loader = clazz.getClassLoader();
        Class<?> serviceClass = null;
        try {
            serviceClass = loader.loadClass(jaxwsWSR.getValue());
        } catch (ClassNotFoundException e) {
            throw new ArchiveInjectionException("Cannot load Service class '" + jaxwsWSR.getValue() + "'", e);
        }

        // Get the wsdllocation
        String wsdlLocation = jaxwsWSR.getWsdlLocation();

        // Transform the WSDL location into an URL
        boolean haveWSDL = true;
        URL location = null;
        if (wsdlLocation != null) {
            try {
                // Try if this is an absolute URL
                location = new URL(wsdlLocation);
            } catch (MalformedURLException mue) {
                // Was not absolute
                // Use the service class's classloader to look up the WSDL URL
                ClassLoader serviceLoader = serviceClass.getClassLoader();
                location = serviceLoader.getResource(wsdlLocation);

                if (location == null) {
                    // OK last chance, try with the TCCL
                    location = loader.getResource(wsdlLocation);
                }

                if (location == null) {
                    // If it's still null at this point, remember it ...
                    haveWSDL = false;
                    // ... and let the JAX-WS Provider load it (JAX-WS 7.2.5)
                }

            }
        } else {
            haveWSDL = false;
        }

        // Get the constructor and create the service instance
        Object service = null;
        try {
            if (haveWSDL) {
                // Use the constructor that can override the WSDL URL value
                Constructor<?> constructor = serviceClass.getConstructor(URL.class, QName.class);
            service = constructor.newInstance(location, null);
            } else {
                // Use the constructor that use default values
                Constructor<?> constructor = serviceClass.getConstructor();
                service = constructor.newInstance();
            }
        } catch (IllegalAccessException e) {
            throw new ArchiveInjectionException("Cannot access the constructor", e);
        } catch (NoSuchMethodException e) {
            throw new ArchiveInjectionException("Unknown constructor", e);
        } catch (InvocationTargetException e) {
            throw new ArchiveInjectionException("Exception when invoking constructor", e);
        } catch (InstantiationException e) {
            throw new ArchiveInjectionException("Cannot create instance", e);
        }

        // By default, return the service instance
        Object injectedValue = service;

        // if type and value are different, then, we have to inject the port, not the service
        if (!jaxwsWSR.getType().equals(jaxwsWSR.getValue())) {
            // SEI interface injection
            ClassLoader serviceClassLoader = serviceClass.getClassLoader();
            Class<?> sei = null;
            try {
                sei = serviceClassLoader.loadClass(jaxwsWSR.getType());
            } catch (ClassNotFoundException e) {
                throw new ArchiveInjectionException("Cannot load SEI '" + jaxwsWSR.getType() + "'", e);
            }

            // Call service.getPort(sei) using reflection (avoid a dependency)
            Method getPortMethod = null;
            try {
                getPortMethod = serviceClass.getMethod("getPort", Class.class);
                injectedValue = getPortMethod.invoke(instance, sei);
            } catch (NoSuchMethodException e) {
                throw new ArchiveInjectionException("getPort(Class<T>) was not found on '" + serviceClass + "'", e);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("getPort(Class<T>) is not accessible on '" + serviceClass + "'", e);
            } catch (InvocationTargetException e) {
                throw new ArchiveInjectionException("Error during execution of '" + getPortMethod + "'", e);
            }

        }

        String encName = jaxwsWSR.getName();

        // Now we have the instance ready to be injected
        // Inject only if metadata is a FieldMetadata
        if (fieldMetadata instanceof ICommonFieldMetadata) {
            setFieldValue((ICommonFieldMetadata) fieldMetadata, clazz, instance, injectedValue);
            encName = buildENCName(jaxwsWSR.getName(), (ICommonFieldMetadata) fieldMetadata);
        }

        // Try without any References/JNDI stuff
        // Direct object in the Context
        // TODO Remember that there is a single shared instance
        bindValue(encName, injectedValue);

    }

    /**
     * @param value tested String
     * @return <code>true</code> if the parameter is null or empty.
     */
    private static boolean isNullOrEmpty(String value) {
        return (value == null) || "".equals(value);
    }


    /**
     * Call postconstruct methods.
     * @param postConstructMethods the methods to call
     * @param clazz the class to use for getting the methods
     * @param instance the given instance to call
     * @throws ArchiveInjectionException if invocation on the methods is not done
     */
    protected void invokePostConstructMethods(List<ICommonMethodMetadata> postConstructMethods, Class clazz, Object instance) throws ArchiveInjectionException {
        // null list, return
        if (postConstructMethods == null) {
            return;
        }

        // Call each method
        for (ICommonMethodMetadata postConstructMethodMetadata : postConstructMethods) {
            Method postConstructMethod = null;
            try {
                postConstructMethod = clazz.getDeclaredMethod(postConstructMethodMetadata.getJMethod().getName());
            } catch (NoSuchMethodException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            }

            boolean accessible = postConstructMethod.isAccessible();
            try {
                postConstructMethod.setAccessible(true);
                postConstructMethod.invoke(instance);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            } catch (InvocationTargetException e) {
                throw new ArchiveInjectionException("Cannot invoke postconstruct method", e);
            } finally {
                postConstructMethod.setAccessible(accessible);
            }
        }

    }


    /**
     * Perform injection in the given instance (or if null on a static way on the class) with the provided metadata
     * @param ejb the EJB-REF that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    public void injectFieldEJB(IJEjbEJB ejb, ICommonFieldMetadata fieldMetadata, Class clazz, Object instance) throws ArchiveInjectionException {
        // get data for the EJB
        String interfaceName = ejb.getBeanInterface();
        String beanName = ejb.getBeanName();
        String mappedName = ejb.getMappedName();

        // Check if there is a value bound in the ENC
        String encName = buildENCName(ejb.getName(), fieldMetadata);
        boolean needToBind = true;
        Object value = null;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }

        // Inject this EJB reference

        // JNDI Name to use if no value yet
        if (value == null) {
            String jndiName = null;

            // Mapped name defined ? use it as JNDI name
            if (mappedName != null) {
                jndiName = mappedName;
            } else {
                // needs to ask the remote side
                jndiName = getRemoteJNDIName(interfaceName, beanName);
            }

            // OK, now that JNDI name is here, get value
            try {
                value = new InitialContext().lookup(jndiName);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + jndiName + "' for ejb '" + ejb + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }

    }

    /**
     * Perform injection in the given instance (or if null on a static way on the class) with the provided metadata
     * @param jAnnotationResource the resource-ref that needs to be injected
     * @param fieldMetadata the metadata that will be used for injection
     * @param instance the instance to use for the injection. May be null for static case.
     */
    public void injectFieldResource(IJAnnotationResource jAnnotationResource, ICommonFieldMetadata fieldMetadata, Class clazz, Object instance) throws ArchiveInjectionException {
        // get data for the Resource
        String mappedName = jAnnotationResource.getMappedName();

        // Value that will be injected
        Object value = null;

        // Check if there is a value bound in the ENC
        String encName = buildENCName(jAnnotationResource.getName(), fieldMetadata);
        boolean needToBind = true;
        try {
            value = new InitialContext().lookup("java:comp/env/" + encName);
            needToBind = false;
        } catch (NamingException e) {
            logger.debug("Value for ENC '{0}' not bound in the registry", encName);
        }


        // No value, needs to find it
        if (value == null) {
            if (mappedName == null) {
                logger.warn("Injection of @Resource for {0} is not supported", jAnnotationResource);
                return;
            }

            // OK, now that JNDI name is here, get value
            try {
                value = new InitialContext().lookup(mappedName);
            } catch (NamingException e) {
                throw new ArchiveInjectionException("Cannot get object with JNDI Name '" + mappedName + "' for Resource '" + jAnnotationResource + "'.", e);
            }
        }

        // Set value
        setFieldValue(fieldMetadata, clazz, instance, value);

        // bind in ENC
        if (needToBind) {
            bindValue(encName, value);
        }

    }

    /**
     * Bind the given value in the ENC.
     * @param encName the ENC name
     * @param value the ENC value
     */
    protected void bindValue(final String encName, Object value) {
        // bind in ENC if not already done
        try {
            new InitialContext().bind("java:comp/env/" + encName, value);
        } catch (NamingException e) {
            logger.warn("Unable to bind the value for ENC '{0}'.", encName);
        }
    }

    /**
     * Sets the value on a given field by using given arguments
     * @param fieldMetadata the metadata for the name of the field
     * @param clazz the class of the field
     * @param instance the instance of the class
     * @param value the value to set
     * @throws ArchiveInjectionException if the value cannot be set
     */
    protected void setFieldValue(ICommonFieldMetadata fieldMetadata, Class clazz, Object instance, Object value) throws ArchiveInjectionException {

        // Value is obtained, set it
        // Get Field from the class
        JField jField = fieldMetadata.getJField();
        Field field = null;
        try {
            field = clazz.getDeclaredField(jField.getName());
        } catch (NoSuchFieldException e) {
            throw new ArchiveInjectionException("Cannot get field '" + jField + "' on the class '" + clazz.getName() + "'", e);
        }

        boolean accessible = field.isAccessible();
        try {
            // needs to be allowed to change value of the field
            field.setAccessible(true);

            // Set value
            try {
                field.set(instance, value);
            } catch (IllegalAccessException e) {
                throw new ArchiveInjectionException("Cannot set given value on the field '" + jField + "'.", e);
            }
        } finally {
            // set back accessible attribute
            field.setAccessible(accessible);
        }

    }

    /**
     * Build an ENC name for a given field.
     * @param name the existing name of the ENC specified through annotation/xml
     * @param fieldMetadata the metadata of the field
     * @return the updated enc name if there was none
     */
    protected String buildENCName(final String name, final ICommonFieldMetadata fieldMetadata) {
        String encName = name;
        if (encName == null || "".equals(encName)) {
            encName = fieldMetadata.getClassMetadata().getJClass().getName().replace("/", ".") + "/"
                    + fieldMetadata.getJField().getName();
        }
        return encName;
    }

    /**
     * Gets a JNDI name from the server for a given bean/interface.
     * @param interfaceName the name of the interface
     * @param beanName the name of the bean
     * @return a JNDI name if value has been found, else an exception
     * @throws ArchiveInjectionException if JNDI Name is not found
     */
    protected String getRemoteJNDIName(String interfaceName, String beanName) throws ArchiveInjectionException {

        String jndiName = null;
        // Init resolver if not yet defined
        if (jndiResolver == null) {
            // Get object
            Object o = null;
            try {
                o = new InitialContext().lookup("EZB_Remote_JNDIResolver");
            } catch (NamingException e) {
                // No Remote JNDI resolver, so throws exception
                throw new ArchiveInjectionException("No Remote EJB3 JNDI Resolver found", e);
            }

            // Object is here, so we can cast it
            jndiResolver = (EZBRemoteJNDIResolver) PortableRemoteObject.narrow(o, EZBRemoteJNDIResolver.class);

        }

        // Ask the Resolver with interface name and bean name
        List<EZBJNDIData> jndiDataList = null;
        try {
            jndiDataList = jndiResolver.getNames(interfaceName, beanName);
        } catch (RemoteException re) {
            // No Remote JNDI resolver, so throw first exception
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "'", re);
        }

        // Data is here, check if it is empty or not
        if (jndiDataList.size() == 0) {
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "', no data was found on the remote side.");
        } else if (jndiDataList.size() > 1) {
            // too many entries
            throw new ArchiveInjectionException("Unable to get JNDI Name for '" + interfaceName + "'/'" + beanName + "', too many answers : '" + jndiDataList + "'.");
        }

        // Only one item found, so get JNDI Name from this object
        EZBJNDIData jndiData = jndiDataList.get(0);
        // Update JNDI name
        jndiName = jndiData.getName();
        if (logger.isDebugEnabled()) {
            logger.debug("Found JNDI Name '" + jndiName + "' for '" + interfaceName + "'/'" + beanName + "', too many answers : '" + jndiDataList + "'.");
        }

        return jndiName;
    }



    /**
     * Gets the metadata for a given class
     * @param clazz the class on which we want to get metadata
     * @return the metadata if found else null
     */
    protected ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> getClassMetadata(Class clazz) {
        return getClassMetadata(clazz.getName());
    }

    /**
     * Gets the metadata for a given class
     * @param requestedClassName the name of the class on which we want to get metadata
     * @return the metadata if found else null
     */
    protected ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> getClassMetadata(final String requestedClassName) {

        // Get an iterator
        Iterator<? extends ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata>> it = metadataCollection.iterator();

        // Search the class by iterating on the metadata collection
        while (it.hasNext()) {
            // Declare temporary var
            ICommonClassMetadata<? extends ICommonClassMetadata, ? extends ICommonMethodMetadata, ? extends ICommonFieldMetadata> tmpMetadata;

            // Get current metadata
            tmpMetadata = it.next();

            // Metadata for the expected class ?
            JClass jClass = tmpMetadata.getJClass();
            if (jClass != null) {
                String className = jClass.getName();
                // compare name
                if (requestedClassName.equals(className.replace('/', '.')))  {
                    return tmpMetadata;
                }

            }
        }

        // Not found
        return null;
    }

}
